/**
 * Collects types for each node.
 * @public
 */
export class TypeMap {
    /**
     * @private
     * @type {Map<string, import('typescript').Type[]>}
     */
    private map;
    /**
     * @param {import('@typescript-eslint/utils').TSESTree.Node} node
     * @param {import('typescript').Type[]} types
     */
    add(node: import("@typescript-eslint/utils").TSESTree.Node, types: import("typescript").Type[]): ts.Type[];
    /**
     * @param {import('@typescript-eslint/utils').TSESTree.Node} node
     */
    get(node: import("@typescript-eslint/utils").TSESTree.Node): ts.Type[];
}
/**
 * Get first element of array or null if empty.
 *
 * @public
 * @template {unknown} T
 * @param {Readonly<T[]>} arr
 * @return {T | null}
 */
export function getFirst<T extends unknown>(arr: Readonly<T[]>): T | null;
/**
 * Get last element of array or null if empty.
 *
 * @public
 * @template {unknown} T
 * @param {Readonly<T[]>} arr
 * @return {T | null}
 */
export function getLast<T extends unknown>(arr: Readonly<T[]>): T | null;
/**
 * Get unique ID for node based on its location.
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 * @returns {string}
 */
export function getNodeID(node: import("@typescript-eslint/utils").TSESTree.Node): string;
/**
 * Get indentation of the node in source code.
 *
 * @public
 * @param {Readonly<import('@typescript-eslint/utils').TSESLint.SourceCode>} sourceCode
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 * @returns {string}
 */
export function getNodeIndent(sourceCode: Readonly<import("@typescript-eslint/utils").TSESLint.SourceCode>, node: import("@typescript-eslint/utils").TSESTree.Node): string;
export const createRule: <Options extends readonly unknown[], MessageIds extends string>({ meta, name, ...rule }: Readonly<ESLintUtils.RuleWithMetaAndName<Options, MessageIds, unknown>>) => ESLintUtils.RuleModule<MessageIds, Options, unknown, ESLintUtils.RuleListener>;
/**
 * @public
 * @param {string} jsdocString
 * @param {string[]} typeStrings
 * @returns {string}
 */
export function appendThrowsTags(jsdocString: string, typeStrings: string[]): string;
/**
 * Check if comment string contains JSDoc throws tag.
 *
 * @public
 * @param {string} comment
 */
export function hasThrowsTag(comment: string): boolean;
/**
 * Check if node has any valid JSDoc.
 *
 * @public
 * @param {Readonly<import('@typescript-eslint/utils').TSESLint.SourceCode>} sourceCode
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 * @return {boolean}
 */
export function hasJSDoc(sourceCode: Readonly<import("@typescript-eslint/utils").TSESLint.SourceCode>, node: import("@typescript-eslint/utils").TSESTree.Node): boolean;
/**
 * Check if node has JSDoc comment with @throws or @exception tag.
 *
 * @public
 * @param {Readonly<import('@typescript-eslint/utils').TSESLint.SourceCode>} sourceCode
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 * @return {boolean}
 */
export function hasJSDocThrowsTag(sourceCode: Readonly<import("@typescript-eslint/utils").TSESLint.SourceCode>, node: import("@typescript-eslint/utils").TSESTree.Node): boolean;
/**
 * Combine multiple types into union type string of given types.
 *
 * @public
 * @param {string[]} typeStrings
 * @return {string}
 */
export function typeStringsToUnionString(typeStrings: string[]): string;
/**
 * Combine multiple types into union type string of given types.
 *
 * @public
 * @param {import('typescript').TypeChecker} checker
 * @param {import('typescript').Type[]} types
 * @return {string}
 */
export function typesToUnionString(checker: import("typescript").TypeChecker, types: import("typescript").Type[]): string;
/**
 * Find closest node that matches the callback predicate.
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node | undefined} node
 * @param {function(import('@typescript-eslint/utils').TSESTree.Node): boolean} callback
 * @returns {import('@typescript-eslint/utils').TSESTree.Node | null}
 */
export function findClosest(node: import("@typescript-eslint/utils").TSESTree.Node | undefined, callback: (arg0: import("@typescript-eslint/utils").TSESTree.Node) => boolean): import("@typescript-eslint/utils").TSESTree.Node | null;
/**
 * Find parent node that matches the callback predicate.
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node | undefined} node
 * @param {function(import('@typescript-eslint/utils').TSESTree.Node): boolean} callback
 * @returns {import('@typescript-eslint/utils').TSESTree.Node | null}
 */
export function findParent(node: import("@typescript-eslint/utils").TSESTree.Node | undefined, callback: (arg0: import("@typescript-eslint/utils").TSESTree.Node) => boolean): import("@typescript-eslint/utils").TSESTree.Node | null;
/**
 * Get call expression node's declaration type.
 *
 * @public
 * @param {import('@typescript-eslint/utils').ParserServicesWithTypeInformation} services
 * @param {import('@typescript-eslint/utils').TSESTree.CallExpression} node
 * @return {import('typescript').Declaration | null}
 */
export function getCallSignatureDeclaration(services: import("@typescript-eslint/utils").ParserServicesWithTypeInformation, node: import("@typescript-eslint/utils").TSESTree.CallExpression): import("typescript").Declaration | null;
/**
 * Get callee node from given node's type.
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 * @return {import('@typescript-eslint/utils').TSESTree.Node | null}
 */
export function getCallee(node: import("@typescript-eslint/utils").TSESTree.Node): import("@typescript-eslint/utils").TSESTree.Node | null;
/**
 * Get all declaration nodes of the callee from the given node's type.
 *
 * @public
 * @param {import('@typescript-eslint/utils').ParserServicesWithTypeInformation} services
 * @param {import('@typescript-eslint/utils').TSESTree.Expression} node
 * @return {import('typescript').Declaration | null}
 */
export function getCalleeDeclaration(services: import("@typescript-eslint/utils").ParserServicesWithTypeInformation, node: import("@typescript-eslint/utils").TSESTree.Expression): import("typescript").Declaration | null;
/**
 * Get all declaration nodes of the callee from the given node's type.
 *
 * @public
 * @param {import('@typescript-eslint/utils').ParserServicesWithTypeInformation} services
 * @param {import('@typescript-eslint/utils').TSESTree.Expression} node
 * @return {import('typescript').Node[]}
 */
export function getCalleeDeclarations(services: import("@typescript-eslint/utils").ParserServicesWithTypeInformation, node: import("@typescript-eslint/utils").TSESTree.Expression): import("typescript").Node[];
/**
 * @public
 * @param {import('typescript').Node} node
 * @returns {Readonly<import('typescript').JSDocThrowsTag[]>}
 */
export function getJSDocThrowsTags(node: import("typescript").Node): Readonly<import("typescript").JSDocThrowsTag[]>;
/**
 * Grab types from only typed throws tags.
 *
 * @public
 * @param {import('typescript').TypeChecker} checker
 * @param {import('typescript').Node} node
 * @returns {import('typescript').Type[]}
 */
export function getJSDocThrowsTagTypes(checker: import("typescript").TypeChecker, node: import("typescript").Node): import("typescript").Type[];
/**
 * Treats union types as separate types.
 *
 * @public
 * @param {import('typescript').Type[]} types
 * @returns {import('typescript').Type[]}
 */
export function toFlattenedTypeArray(types: import("typescript").Type[]): import("typescript").Type[];
/**
 * Collect function call expression nodes for given identifier node.
 *
 * @public
 * @param {Readonly<import('@typescript-eslint/utils').TSESLint.SourceCode>} sourceCode
 * @param {import('@typescript-eslint/utils').TSESTree.Identifier} node
 * @return {import('@typescript-eslint/utils').TSESTree.CallExpression[]}
 */
export function findFunctionCallNodes(sourceCode: Readonly<import("@typescript-eslint/utils").TSESLint.SourceCode>, node: import("@typescript-eslint/utils").TSESTree.Identifier): import("@typescript-eslint/utils").TSESTree.CallExpression[];
/**
 * Find closest function where exception is thrown
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 * @returns {import('@typescript-eslint/utils').TSESTree.FunctionLike | null}
 */
export function findClosestFunctionNode(node: import("@typescript-eslint/utils").TSESTree.Node): import("@typescript-eslint/utils").TSESTree.FunctionLike | null;
/**
 * Find where JSDoc comment should be added
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.FunctionLike | import('@typescript-eslint/utils').TSESTree.Identifier} node
 * @returns {import('@typescript-eslint/utils').TSESTree.Node | null}
 */
export function findNodeToComment(node: import("@typescript-eslint/utils").TSESTree.FunctionLike | import("@typescript-eslint/utils").TSESTree.Identifier): import("@typescript-eslint/utils").TSESTree.Node | null;
/**
 * Find declaration node of identifier node
 *
 * @public
 * @param {Readonly<import('@typescript-eslint/utils').TSESLint.SourceCode>} sourceCode
 * @param {import('@typescript-eslint/utils').TSESTree.Identifier} node
 * @return {import('@typescript-eslint/utils').TSESTree.Node | null}
 */
export function findIdentifierDeclaration(sourceCode: Readonly<import("@typescript-eslint/utils").TSESLint.SourceCode>, node: import("@typescript-eslint/utils").TSESTree.Identifier): import("@typescript-eslint/utils").TSESTree.Node | null;
/**
 * Check if node is in try-catch block where exception is handled
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node | undefined} node
 * @returns {boolean}
 */
export function isInHandledContext(node: import("@typescript-eslint/utils").TSESTree.Node | undefined): boolean;
/**
 * Check if node promise rejection handled.
 * Such as `try .. await node .. catch` or `node.catch(...)`
 *
 * @public
 * @param {Readonly<import('@typescript-eslint/utils').TSESLint.SourceCode>} sourceCode
 * @param {import('@typescript-eslint/utils').TSESTree.Node | undefined} node
 * @returns {boolean}
 */
export function isInAsyncHandledContext(sourceCode: Readonly<import("@typescript-eslint/utils").TSESLint.SourceCode>, node: import("@typescript-eslint/utils").TSESTree.Node | undefined): boolean;
/**
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 */
export function isNodeReturned(node: import("@typescript-eslint/utils").TSESTree.Node): boolean;
/**
 * Check if type is exactly a Promise type.
 *
 * @public
 * @param {import('@typescript-eslint/utils').ParserServicesWithTypeInformation} services
 * @param {import('typescript').Type} type
 * @returns {boolean}
 */
export function isPromiseType(services: import("@typescript-eslint/utils").ParserServicesWithTypeInformation, type: import("typescript").Type): boolean;
/**
 * Check if node is a callback function of Promise constructor.
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 */
export function isPromiseConstructorCallbackNode(node: import("@typescript-eslint/utils").TSESTree.Node): boolean;
/**
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 * @returns {node is import('@typescript-eslint/utils').TSESTree.MethodDefinition | import('@typescript-eslint/utils').TSESTree.Property}
 */
export function isAccessorNode(node: import("@typescript-eslint/utils").TSESTree.Node): node is import("@typescript-eslint/utils").TSESTree.MethodDefinition | import("@typescript-eslint/utils").TSESTree.Property;
/**
 * Check if node is a callback function of one of the promise chain thenable methods.
 *
 * @public
 * @param {import('@typescript-eslint/utils').TSESTree.Node} node
 */
export function isThenableCallbackNode(node: import("@typescript-eslint/utils").TSESTree.Node): boolean;
import ts = require("typescript");
import { ESLintUtils } from "@typescript-eslint/utils";
//# sourceMappingURL=utils.d.ts.map